;;Sets any environment variables and loads ml4pg and sepia menus
(defun coqpr3-mode ()
  (defvar coqpr3location nil)
  (setq coqpr3location (symbol-file 'coqpr3-mode))
  (setq coqpr3home (file-name-directory coqpr3location))
  (if (not (file-exists-p "~/coqpr3")) (make-directory "~/coqpr3") nil)
  (setenv "ML4PG_HOME" (concat coqpr3home "ml4pg/"))
  (load-file (concat coqpr3home "sepia/efsm.el"))
  (load-file (concat coqpr3home "ml4pg/ml4pg.el"))
  (initialise-coqpr3))

;;Adds any menus and asks the user which ML4PG mode they are using
(defun initialise-coqpr3 ()
  (addmenus)
  (addcoqpr3menus)
  (select-mode))

;;Adds the menus for CoqPR3 specific functionality
(defun addcoqpr3menus ()
  (easy-menu-define coqpr3-menu global-map "CoqPR3"
    '("CoqPR3"
      ["Step 0: Extract and save data from other theories" (auto-extract)]
      ["Step 1: Extract info upto lemma" (progn (extract-feature-theorems) (output-hyps))]
      ["Step 2: Identify similar lemmas" (get-similar)]
      ["Step 3: Generate clusters for selected libraries" (progn (show-clusters-bis) (write-region nil nil "tmp_clusters"))]
      ["Step 4: Invoke model inference" (proof-setup-coqpr3)]
      ))
  (easy-menu-remove-item global-map '("menu-bar") "CoqPR3")
  (easy-menu-add-item nil nil coqpr3-menu "help-menu"))

  ;;Automatic extraction for theories
  (defun auto-extract ()
  (progn
  (goto-char (point-max))
  (re-search-backward "Qed")
  (end-of-line)
  (extract-feature-theorems)
  (save-numbers)
  (write-region nil nil (concat "~/coqpr3/" theoryname ".v"))
  (kill-emacs)))

;;Invokes ML4PG to generate similarities for current proof goal and saves to a temp file
(defun get-similar()
  (progn
  (show-clusters-of-theorem)
  (write-region nil nil "tmp_suggestions")
  (switch-to-previous-buffer)))

;;Outputs the hypotheses for lemmas appearing before the current proof goal and saves to temp file
(defun output-hyps()
  (set-hypotheses-file "tmp_hyps")
  (write-hypotheses))

;;Whenever a Coq file is opened, invoke CoqPR3 mode
(add-hook 'coq-mode-hook 'coqpr3-mode)

;;A function to return to the previous buffer
(defun switch-to-previous-buffer ()
  "Switch to previously open buffer.
Repeated invocations toggle between the two most recently open buffers."
  (interactive)
  (switch-to-buffer (other-buffer (current-buffer) 1)))

;;Code that actually invokes SEPIA for a proof attempt
  (defun invoke-sepia ()
    (save-buffer)
    (setq maintheory (buffer-file-name))
    (get-loaded-theories (buffer-list))
    (setq currentbuffer (buffer-name))
        (with-output-to-temp-buffer "*sepia-output*"
      (async-shell-command (concat "java"
                             " -cp "
  			   (eval 'coqpr3home)
           "sepia/EFSMTool-1.0-SNAPSHOT-jar-with-dependencies.jar"
  			   " org.bitbucket.efsmtool.proofs.utilities.CoqPR3Entry"
           " -lemmastatement "
  			   (concat "\""  thisline "\"")
  			   " -maintheory "
  			   maintheory
  			   " -additionaltheories "
  			   othertheories
  			   " -limit  "
                 	(eval 'limit)
                 	" -timeout "
                 	(eval 'timeout)
                  " -strategy "
                  (eval 'strategy)
                   " -k "
                 	 (eval 'k)
                 	 " -lineintheory "
                 	 (concat "\"" (eval 'currentline) "\""))
                     "*sepia-output*"
                     "*sepia-output*")
      (pop-to-buffer "*sepia-output*"))
      (highlight-lines-matching-regexp "^.*Proof was:.*$" 'hi-green-b)
            (highlight-lines-matching-regexp "^.*Exhausted.*$" 'hi-red-b))

            (defun proof-setup-coqpr3 ()
              (point)
              (re-search-backward "Lemma\\|Theorem")
              (setq currentline (what-line))
              (if (or (looking-at "Lemma.*\\.$") (looking-at "Theorem.*\\.$"))
                  (progn
            	(setq p1 (line-beginning-position))
            	(setq p2 (line-end-position))
            	(setq thisline (s-trim (buffer-substring-no-properties p1 p2))))
            	(progn
            		(setq p1 (line-beginning-position))
                	(setq p2 (line-end-position))
                	(setq thisline (s-trim (buffer-substring-no-properties p1 p2)))
                (while (progn
                    	    (forward-line 1)
                (setq p1 (line-beginning-position) )
                	    (setq p2 (line-end-position) )
                	    (setq subsline (s-trim (buffer-substring-no-properties p1 p2)))
                	    (setq thisline (concat thisline " " subsline " "))
                	    (not (looking-at ".*\\.$"))))
                )
            )
            (message thisline)
            (setq workingdirectory (file-name-directory (buffer-file-name)))
              (invoke-sepia))

;;A function to store the name of the main theory
(defun store-theory-name ()
  (setq theoryname (file-name-base (buffer-file-name)))
  (message (concat "Working with theory: " theoryname)))
