(defvar SEPIAdir "~/Research/software/sepia/efsminferencetool/EFSMTool/proofgeneral-integration/")
(defvar limit           "nil")
(defvar strategy   "redblue")
(defvar k               "1")
(defvar timeout "30")
(defvar search "breadth")

(defun change-limit (n)
  (setq limit n)
  (setq timeout "nil")
  (message limit)
  (message timeout))

(defun change-timeout (n)
  (setq timeout n)
  (setq limit "nil")
  (message timeout)
  (message limit))

(defun change-strategy (s)
  (setq strategy s))

(defun change-search (se)
(setq search se))

(defun change-k (n)
  (setq k n))

(defun initialise-sepia ()
  (addmenus))

(defun custom-limit ()
  (interactive)
  (let* ((msg   "Please enter a number of tactics to apply: ")
         (smode (if noninteractive "" (read-string msg))))
         (setq limit smode)
         (setq timeout "nil"))
  (message limit))

(defun custom-timeout ()
  (interactive)
  (let* ((msg   "Please enter a timeout (in seconds e.g. 60s): ")
         (smode (if noninteractive "" (read-string msg))))
         (setq timeout smode)
         (setq limit "nil"))
         (message timeout))

(require 'cl)

(defun addmenus ()
  (easy-menu-define efsm-menu global-map "SEPIA"
    '("SEPIA"
      ["Prove Current Lemma" (proof-setup)]
      ["Analyse Proofs" (analyse-proofs)]
      ("Search Type"
      ["depth" (change-search "depth")]
      ["breadth" (change-search "breadth")]
      ["best - less subgoals" (change-search "best-lesssubgoals")]
      ["best - small context" (change-search "best-smallcontext")]
      ["best - large context" (change-search "best-largecontext")])
      ("MINT Options"
       ("Inference Strategy"
	["redblue" (change-strategy "redblue")]
	["noloops" (change-strategy "noloops")]
	["exhaustive" (change-strategy "exhaustive")])
       ("K-Value"
	["0" (change-k "0")]
	["1" (change-k "1")]
	["2" (change-k "2")]))
      ("Search Limits"
       ["1,000" (change-limit "1000")]
     ["10,000" (change-limit "10000")]
     ["25,000" (change-limit "25000")]
     ["50,000" (change-limit "50000")]
     ["100,000" (change-limit "100000")]
     ["custom" (custom-limit)])
    ("Timeout"
	["5s" (change-timeout "5")]
	["10s" (change-timeout "10")]
	["30s" (change-timeout "30")]
	["60s" (change-timeout "60")]
	["custom" (custom-timeout)])))
  (easy-menu-remove-item global-map '("menu-bar") "SEPIA")
  (easy-menu-add-item nil nil efsm-menu "help-menu"))

(defun proof-setup ()
  (point)
  (re-search-backward "Lemma\\|Theorem")
  (setq currentline (what-line))
  (if (or (looking-at "Lemma.*\\.$") (looking-at "Theorem.*\\.$"))
      (progn
	(setq p1 (line-beginning-position))
	(setq p2 (line-end-position))
	(setq thisline (s-trim (buffer-substring-no-properties p1 p2))))
	(progn
		(setq p1 (line-beginning-position))
    	(setq p2 (line-end-position))
    	(setq thisline (s-trim (buffer-substring-no-properties p1 p2)))
    (while (progn
        	    (forward-line 1)
    (setq p1 (line-beginning-position) )
    	    (setq p2 (line-end-position) )
    	    (setq subsline (s-trim (buffer-substring-no-properties p1 p2)))
    	    (setq thisline (concat thisline " " subsline " "))
    	    (not (looking-at ".*\\.$"))))
    )
)
(message thisline)
  (communicate-with-sepia-jar))

(defun communicate-with-sepia-jar ()
  (save-buffer)
  (setq maintheory (buffer-file-name))
  (get-loaded-theories (buffer-list))
  (setq currentbuffer (buffer-name))
      (with-output-to-temp-buffer "*sepia-output*"
    (async-shell-command (concat "java"
                           " -cp "
			   SEPIAdir
			   "EFSMTool-1.0-SNAPSHOT-jar-with-dependencies.jar"
			   " org.bitbucket.efsmtool.proofs.utilities.ProofGeneralEntry"
			   " -lemmastatement "
			   (concat "\""  thisline "\"")
			   " -maintheory "
			   maintheory
			   " -additionaltheories "
			   othertheories
			   " -limit  "
               	(eval 'limit)
               	" -timeout "
               	(eval 'timeout)
                " -strategy "
                (eval 'strategy)
                 " -k "
               	 (eval 'k)
               	 " -lineintheory "
               	 (concat "\"" (eval 'currentline) "\"")
                           )
                   "*sepia-output*"
                   "*sepia-output*")
    (pop-to-buffer "*sepia-output*"))
    (highlight-lines-matching-regexp "^.*Proof was:.*$" 'hi-green-b)
          (highlight-lines-matching-regexp "^.*Exhausted.*$" 'hi-red-b))

(defun get-loaded-theories (list)
  (setq othertheories "")
  (while list
    (setq elem (buffer-file-name (car list)))
    (if (string= elem "nil")
	(setq list (cdr list))
      (when (string-match ".*\.v$" elem)
	(if (string= elem maintheory)
	    ()
	  (setq othertheories (concat othertheories " " elem))))
      (setq list (cdr list)))))




      (defun s-trim-left (s)
  (if (string-match "\\`[ \t\n\r]+" s)
      (replace-match "" t t s)
    s))

(defun s-trim-right (s)
  (if (string-match "[ \t\n\r]+\\'" s)
      (replace-match "" t t s)
    s))

(defun s-trim (s)
  (s-trim-left (s-trim-right s)))

(defun analyse-proofs ()
    (setq maintheory (buffer-file-name))
       (with-output-to-temp-buffer "*sepia-output*"
    (async-shell-command (concat "java"
                           " -cp "
			   SEPIAdir
			   "EFSMTool-1.0-SNAPSHOT-jar-with-dependencies.jar"
			   " org.bitbucket.efsmtool.proofs.utilities.ProofAnalyser"
			   " -folds 10 "
			   " -theory "
			   maintheory
			   " -limit  "
			   (eval 'limit)
			   " -timeout "
			   (eval 'timeout)
			   " -strategy "
			   (eval 'strategy)
			   " -k "
			   (eval 'k)
                           )
                   "*sepia-output*"
                   "*sepia-output*")
    (pop-to-buffer "*sepia-output*")))

(defun sepia-mode ()
  (initialise-sepia))


(add-hook 'coq-mode-hook 'sepia-mode)
